/*
Copyright 2017 Detelina Stoyanova

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
 */

package forAge;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.zip.DataFormatException;

/**
 * This function creates all the instances of different classes and calculates the age estimates give the scores from the different algorithms
 * 
 * @author Detelina Stoyanova
 * @version 20180217
 * 
 */

public class ForAge 
{
	private static BoneScan boneScan = null;
	private static Positioning position;
	private static SAH sahScore;
	private static Mesh meshTPS;
	private static TPS tps;
	private static Mesh meshCurv;
	private static VentralCurvature vc;
	
	// the age estimates using each of the individual scores
	private double BEageEstimate;
	private double SAHageEstimate;
	private double VCageEstimate;
	
	// the age estimates from the multivariate regression models
	private double BEandVCageEstimate;
	private double SAHandVCageEstimate;
	
	public ForAge(String fileText) throws FileNotFoundException, IOException, DataFormatException
	{
		start(fileText);
	}
	
	public void start(String fileText) throws FileNotFoundException, IOException, DataFormatException
	{
		boneScan = new BoneScan(fileText);
		
		sahScore = new SAH(boneScan.getNumVertices(), boneScan.getX(), boneScan.getY(), boneScan.getZ());
		
		position = new Positioning(boneScan.getNumVertices(), boneScan.getX(), boneScan.getY(), boneScan.getZ());
		
		boneScan.updateCoordinates(position.getX(), position.getY(), position.getZ());
	}
	
	public void doCalculations()
	{
		meshTPS = new Mesh(21, boneScan.getNumFaces(), boneScan.getX(), boneScan.getY(), boneScan.getZ(), boneScan.getV1(), boneScan.getV2(), boneScan.getV3(), position.getMaxX(), position.getMaxY(), position.getMinX(), position.getMinY(), position.getMaxZ(), position.getMinZ());
		
		tps = new TPS(meshTPS.getMeshTotalPts(), meshTPS.getX(), meshTPS.getY(), meshTPS.getScanPtsZ());
		
		meshCurv = new Mesh(61, boneScan.getNumFaces(), boneScan.getX(), boneScan.getY(), boneScan.getZ(), boneScan.getV1(), boneScan.getV2(), boneScan.getV3(), position.getMaxX(), position.getMaxY(), position.getMinX(), position.getMinY(), position.getMaxZ(), position.getMinZ());
		
		vc = new VentralCurvature(meshCurv.getNumOutlinePts(), meshCurv.getOutlinePtsX(), meshCurv.getOutlinePtsY());
		
		calculateBEEstimate(tps.getBE());
		calculateSAHScoreEstimate(sahScore.getSAH());
		calculateVentralCurvatureEstimate(vc.getVentralCurvatureValue());
		
		System.out.println("Curvature = " + vc.getVentralCurvatureValue());
		System.out.println("SAH = " + sahScore.getSAH());
		System.out.println("BE = " + tps.getBE());
		
		calculateBEandVCEstimate(tps.getBE(), vc.getVentralCurvatureValue());
		calculateSAHandVCEstimate(sahScore.getSAH(), vc.getVentralCurvatureValue());
	}
	
	// calculates the age estimate from the BE measure using the coefficients of the linear regression
	private void calculateBEEstimate(double be)
	{	
		double a = -0.4069;
		double b = 3.6024;
		
		double logest = a*Math.log(be)+b;
		
		BEageEstimate = Math.exp(logest);
	}
	
	// calculates the age estimate from the SAH-Score using the coefficients of the linear regression
	private void calculateSAHScoreEstimate(double var)
	{				
		double a = -0.8496;
		double b = 2.0595;
		
		double logest = a * Math.log(var) + b;
		
		SAHageEstimate = Math.exp(logest);
	}
	
	// calculates the age estimate from the Ventral Curvature measure using the coefficients of the linear regression
	private void calculateVentralCurvatureEstimate(double curv)
	{
		double a = 6.7344;
		double b = 3.0739;
		
		double logest = a * curv + b;
		
		VCageEstimate = Math.exp(logest);
	}
	
	// calculates the age estimate from the BE measure and the Ventral Curvature measure using the coefficients of the multivariate regression model
	private void calculateBEandVCEstimate(double be, double curv)
	{
		double logest = 3.1805 + 4.8460*curv - 0.3312*Math.log(be);
		BEandVCageEstimate = Math.exp(logest);
	}
	
	// calculates the age estimate from the SAH-Score and the Ventral Curvature measure using the coefficients of the multivariate regression model
	private void calculateSAHandVCEstimate(double sah, double curv)
	{
		double logest = 2.015 + 3.0674*curv - 0.7281*Math.log(sah);
		SAHandVCageEstimate = Math.exp(logest);
	}
	
	public BoneScan getBoneScan()
	{
		return boneScan;
	}
	
	public double getBEAge()
	{
		return BEageEstimate;
	}
	
	public double getSAHAge()
	{
		return SAHageEstimate;
	}
	
	public double getVCAge()
	{
		return VCageEstimate;
	}
	
	public double getBE()
	{
		return tps.getBE(); 
	}
	
	public double getSAH()
	{
		return sahScore.getSAH();
	}
	
	public double getVC()
	{
		return vc.getVentralCurvatureValue();
	}
	
	public double getBEandVCAge()
	{
		return BEandVCageEstimate;
	}
	
	public double getSAHandVCAge()
	{
		return SAHandVCageEstimate;
	}
	
	public String getCurvatureMeshString()
	{	
		String initMesh = "";

		for (int i = 0; i < meshCurv.getNumOutlinePts(); i++)
		{
			initMesh += "v " + meshCurv.getOutlinePtsX(i) + " " + meshCurv.getOutlinePtsY(i) + " " + meshCurv.getOutlinePtsZ(i) + "\n";
		}
		
		return initMesh;
	}
	
	public float [][] getInitMeshArray()
	{
		float [][] initMesh = new float[meshTPS.getInitMeshTotalPts()][3]; 
		
		for (int i = 0; i < meshTPS.getInitMeshTotalPts(); i++)
		{
			initMesh[i][0] = (float)meshTPS.getInitX(i);
			initMesh[i][1] = (float)meshTPS.getInitY(i);
			initMesh[i][2] = (float)meshTPS.getInitZ(i);
		}
		
		return initMesh;
	}
	
	public String getTPSMeshString()
	{
		String finalMesh = "";

		for (int i = 0; i < meshTPS.getMeshTotalPts(); i++)
		{
			finalMesh += "v " + meshTPS.getScanPtsX(i) + " " + meshTPS.getScanPtsY(i) + " " + meshTPS.getScanPtsZ(i) + "\n";
		}
		
		return finalMesh;
	}
	
	public float [][] getTPSMeshArray()
	{
		float [][] finalMesh = new float[meshTPS.getMeshTotalPts()][3]; 
		
		for (int i = 0; i < meshTPS.getMeshTotalPts(); i++)
		{
			finalMesh[i][0] = (float)meshTPS.getScanPtsX(i);
			finalMesh[i][1] = (float)meshTPS.getScanPtsY(i);
			finalMesh[i][2] = (float)meshTPS.getScanPtsZ(i);
		}
		
		return finalMesh;
	}
	
	public float [][] getCurvatureMeshArray()
	{
		float [][] curvatureMesh = new float[meshCurv.getNumOutlinePts()][3]; 
		
		for (int i = 0; i < meshCurv.getNumOutlinePts(); i++)
		{
			curvatureMesh[i][0] = (float)meshCurv.getOutlinePtsX(i);
			curvatureMesh[i][1] = (float)meshCurv.getOutlinePtsY(i);
			curvatureMesh[i][2] = (float)meshCurv.getOutlinePtsZ(i);
		}
		
		return curvatureMesh;
	}
	
	public String getPCAString()
	{
		String pcastr = "";
		
		String header = "";
		header += "ply\nformat ascii 1.0\nelement vertex ";
		header += (boneScan.getNumVertices() + 1) + "\n";
		header += "property float x\nproperty float y\nproperty float z\nelement face ";
		header += boneScan.getNumFaces() + "\n";
		header += "property list uchar int vertex_indices\nend_header\n0 0 0\n";

		pcastr += header;
		for(int i = 0; i < position.getNumVer(); i++)
		{
			pcastr += position.getX(i) + " " + position.getY(i) + " " + position.getZ(i) + "\n";
		}
		for (int i = 0; i < boneScan.getNumFaces(); i++)
		{
			pcastr += "3 " + (boneScan.getV1(i)+1) + " " + (boneScan.getV2(i)+1) + " " + (boneScan.getV3(i)+1) + "\n";
		}
		
		return pcastr;
	}
	
	public float [][] getPCAVerticesArray()
	{
		float [][] pcaVertices = new float[position.getNumVer()][3]; 
		
		for (int i = 0; i < position.getNumVer(); i++)
		{
			pcaVertices[i][0] = (float)position.getX(i);
			pcaVertices[i][1] = (float)position.getY(i);
			pcaVertices[i][2] = (float)position.getZ(i);
		}
		
		return pcaVertices;
	}
	
	public int [][] getPCAFacesArray()
	{
		int [][] pcaFaces = new int[boneScan.getNumFaces()][3]; 
		
		for (int i = 0; i < boneScan.getNumFaces(); i++)
		{
			pcaFaces[i][0] = boneScan.getV1(i);
			pcaFaces[i][1] = boneScan.getV2(i);
			pcaFaces[i][2] = boneScan.getV3(i);
		}
		
		return pcaFaces;
	}
}
