/*
Copyright 2017 Detelina Stoyanova

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
 */

package forAge;
import Jama.*; 

/**
 * This function applies affine transformations (translation, scaling and rotation) to the scan data
 * 
 * @author Detelina Stoyanova
 * @version 20180217
 * 
 */
public class Positioning 
{
	// the coordinates of the scan
	private double [] x;
	private double [] y;
	private double [] z;
	
	// the number of vertices on the scan
	private int numVer;
	
	// the min and max values of the coordinates on the scan in each dimension
	private double maxX, maxY, maxZ;
	private double minX, minY, minZ;
	
	// the center of the scan
	private double centerX, centerY, centerZ;
	
	public Positioning( int an, double [] ax, double [] ay, double [] az )
	{
		numVer = an;
		
		x = new double[numVer];
		y = new double[numVer];
		z = new double[numVer];
		
		x = ax;
		y = ay;
		z = az;
		
		translate();
		PCA();
		translate();
		scale();
		translate();
	}
	
	// affine translation of the center of the scan to the center of the coordinate system
	private void translate()
	{
		findAverageCenter();
		
		for (int i = 0; i < numVer; i++)
		{
			x[i] -= centerX;
			y[i] -= centerY;
			z[i] -= centerZ;
		}
	}
	
	// PCA (Principal Component Analysis) rotation
	private void PCA()
	{
		double [][] array = new double[numVer][3];
		for (int i = 0; i < numVer; i ++)
		{
			array[i][0] = x[i];
			array[i][1] = y[i];
			array[i][2] = z[i];
		}
		
		Matrix A = new Matrix(array);
		Matrix At = A.transpose();
		Matrix AtA = At.times(A);
		
		Matrix V = AtA.svd().getV();
		
		Matrix res = A.times(V);
		
		double [][] resArray = new double[numVer][3];
		resArray = res.getArray();
		
		for (int i = 0; i < numVer; i ++)
		{
			x[i] = resArray[i][0];
			y[i] = resArray[i][1];
			z[i] = resArray[i][2];
		}
	}
	
	// scaling of the data
	private void scale()
	{
		double dx0 = maxX - minX;
		double dx1 = 10.0; // final scan length;

		double scaleFactor = dx1/dx0;

		for (int i = 0; i < numVer; i++) 
		{
			x[i] *= scaleFactor;
			y[i] *= scaleFactor; 
			z[i] *= scaleFactor;
		} 
	}
	
	// finding the center of the scan as the mean value of all coordinates
	public void findMeanCenter()
	{
		centerX = 0;
		centerY = 0;
		centerZ = 0;

		for ( int i = 0; i < numVer; i++)
		{
			centerX += x[i];
			centerY += y[i];
			centerZ += z[i];
		}

		centerX /= numVer;
		centerY /= numVer;
		centerZ /= numVer;
	}
	
	//finding the center of the scan as the average value of the min and max coordinates in each dimension
	public void findAverageCenter()
	{
		findMinMax();

		centerX = (maxX + minX)/2.0;
		centerY = (maxY + minY)/2.0;
		centerZ = (maxZ + minZ)/2.0;
	}
	
	// find the min and max coordinates in each dimension
	private void findMinMax()
	{
		maxX = x[0];
		maxY = y[0];
		maxZ = z[0];

		for (int i = 1; i < numVer; i++)
		{
			if (x[i] > maxX)
			{
				maxX = x[i];
			}
			if (y[i] > maxY)
			{
				maxY = y[i];
			}
			if (z[i] > maxZ)
			{
				maxZ = z[i];
			}
		}

		minX = x[0];
		minY = y[0];
		minZ = z[0];

		for (int i = 1; i < numVer; i++)
		{
			if (x[i] < minX)
			{
				minX = x[i];
			}
			if (y[i] < minY)
			{
				minY = y[i];
			}
			if (z[i] < minZ)
			{
				minZ = z[i];
			}
		}
	}

	public int getNumVer()
	{
		return numVer;
	}
	
	public double[] getX()
	{
		return x;
	}
	
	public double[] getY()
	{
		return y;
	}
	
	public double[] getZ()
	{
		return z;
	}
	
	public double getX(int index)
	{
		return x[index];
	}
	
	public double getY(int index)
	{
		return y[index];
	}
	
	public double getZ(int index)
	{
		return z[index];
	}
	
	public double getCenterX()
	{
		return centerX;
	}
	
	public double getCenterY()
	{
		return centerY;
	}
	
	public double getCenterZ()
	{
		return centerZ;
	}
	
	public double getMaxX()
	{
		return maxX;
	}
	
	public double getMaxY()
	{
		return maxY;
	}
	
	public double getMaxZ()
	{
		return maxZ;
	}
	
	public double getMinX()
	{
		return minX;
	}
	
	public double getMinY()
	{
		return minY;
	}
	
	public double getMinZ()
	{
		return minZ;
	}
}
