/*
Copyright 2017 Detelina Stoyanova

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
 */

package forAge;

import Jama.*; 

/**
 * This function calculates the SAH-Score
 * 
 * @author Detelina Stoyanova
 * @version 20180217
 * 
 */
public class SAH 
{
	// the x, y and z coordinates of the scan
	private double [] x;
	private double [] y;
	private double [] z;
	
	// the number of vertices on the scan
	private int numVer;
	
	// the center of the scan
	private double centerX, centerY, centerZ;
	
	// the min and max values of the coordinates on the scan in each dimension
	private double maxX, maxY, maxZ;
	private double minX, minY, minZ;
	
	
	private double SAHScore;
	
	public SAH(int an, double [] ax, double [] ay, double [] az)
	{
		numVer = an;
		
		x = new double[numVer];
		y = new double[numVer];
		z = new double[numVer];
		
		x = ax;
		y = ay;
		z = az;

		translate();
		PCA();
		findMinMax();
		scale();
		calculateSAHScore();
	}
	
	// scaling of the data
	private void scale()
	{
		double dx0 = maxX - minX;
		double dx1 = 10.0; // final scan length;

		double scaleFactor = dx1/dx0;

		for (int i = 0; i < numVer; i++) 
		{
			x[i] *= scaleFactor;
			y[i] *= scaleFactor; 
			z[i] *= scaleFactor;
		} 
	}
	
	// translation to the center of the coordinate system using the mean center
	private void translate()
	{
		findMeanCenter();
		
		for (int i = 0; i < numVer; i++)
		{
			x[i] -= centerX;
			y[i] -= centerY;
			z[i] -= centerZ;
		}

		findMinMax();
	}
	
	// PCA (Principal Component Analysis) rotation
	private void PCA()
	{
		double [][] array = new double[numVer][3];
		for (int i = 0; i < numVer; i ++)
		{
			array[i][0] = x[i];
			array[i][1] = y[i];
			array[i][2] = z[i];
		}
		
		Matrix A = new Matrix(array);
		Matrix At = A.transpose();
		Matrix AtA = At.times(A);
		
		Matrix V = AtA.svd().getV();

		Matrix AV = A.times(V);
		
		double [][] av = AV.getArray();
		
		for( int i = 0; i < numVer; i++)
		{
			x[i] = av[i][0];
			y[i] = av[i][1];
			z[i] = av[i][2];
		}
	}
	
	// calculating the SAH-Score as the third singular value 
	private void calculateSAHScore()
	{
		double [][] array = new double[numVer][3];
		for (int i = 0; i < numVer; i ++)
		{
			array[i][0] = x[i];
			array[i][1] = y[i];
			array[i][2] = z[i];
		}
		
		Matrix A = new Matrix(array);
		Matrix At = A.transpose();
		Matrix AtA1 = At.times(A);
		
		double [][] ata = AtA1.getArray();
		for ( int i = 0; i < 3; i ++ )
		{
			for ( int j = 0; j < 3; j ++ )
			{
				ata[i][j] /= (numVer - 1.0); 
			}	
		}
		
		Matrix AtA = new Matrix(ata);
		
		Matrix S = AtA.svd().getS();

		double [][] s = S.getArray();
		
		
		SAHScore = Math.sqrt(s[2][2]);
		
	}
	
	// finding the center of the scan as the mean value of all coordinates
	public void findMeanCenter()
	{
		centerX = 0;
		centerY = 0;
		centerZ = 0;

		for ( int i = 0; i < numVer; i++)
		{
			centerX += x[i];
			centerY += y[i];
			centerZ += z[i];
		}

		centerX /= numVer;
		centerY /= numVer;
		centerZ /= numVer;
	}
	
	public double getSAH()
	{
		return SAHScore;
	}
	
	// find the min and max coordinates in each dimension
	private void findMinMax()
	{
		maxX = x[0];
		maxY = y[0];
		maxZ = z[0];

		for (int i = 1; i < numVer; i++)
		{
			if (x[i] > maxX)
			{
				maxX = x[i];
			}
			if (y[i] > maxY)
			{
				maxY = y[i];
			}
			if (z[i] > maxZ)
			{
				maxZ = z[i];
			}
		}

		minX = x[0];
		minY = y[0];
		minZ = z[0];

		for (int i = 1; i < numVer; i++)
		{
			if (x[i] < minX)
			{
				minX = x[i];
			}
			if (y[i] < minY)
			{
				minY = y[i];
			}
			if (z[i] < minZ)
			{
				minZ = z[i];
			}
		}
	}
}
