/*
Copyright 2017 Detelina Stoyanova

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
 */

package forAge;

import Jama.*; 

/**
 * This function implements the TPS algorithm and calculates the minimum BE
 * 
 * @author Detelina Stoyanova
 * @version 20180217
 * 
 */
public class TPS 
{
	private double [] controlSet1X;
	private double [] controlSet1Y;
	
	private double [] controlSet2Z;
	
	private int numPoints;
	
	private double BE;
	
	// we represent the z coordinaes as a function of x and y coordinates
	public TPS(int an, double[] aC1x, double[] aC1y, double[] aC2z)
	{
		controlSet1X = aC1x;
		controlSet1Y = aC1y;

		controlSet2Z = aC2z;
		
		numPoints = an;
		
		tps();
	}

	// the implementation of the Thin Plate Splines algorithm
	// for details of the algorithm refer to Stoyanova et al., 2015
	// http://onlinelibrary.wiley.com/doi/10.1002/ajpa.22797/full
	private void tps()
	{
		double [][] P = new double[numPoints][3];
		for ( int i = 0; i < numPoints; i++)
		{
			P[i][0] = 1;
			P[i][1] = controlSet1X[i];
			P[i][2] = controlSet1Y[i];
		}

		double temp1, temp2, temp3;
		double basis_fun;

		double [][] C = new double[numPoints][numPoints];

		for ( int i = 0; i < numPoints; i ++ )
		{
			for ( int j = 0; j < numPoints; j ++ )
			{
				temp1 = controlSet1X[i] - controlSet1X[j];
				temp2 = controlSet1Y[i] - controlSet1Y[j];

				temp3 = Math.sqrt(temp1*temp1 + temp2*temp2);

				basis_fun = basisFunct( temp3 );

				C[i][j] = basis_fun;		
			}	
		} 

		// create the matrix A
		double [][] A = new double[numPoints+3][numPoints+3];
		for ( int i = 0; i < numPoints; i++)
		{
			for ( int j = 0; j < numPoints; j++)
			{
				A[i][j] = C[i][j];
			}
			A[i][numPoints] = P[i][0];
			A[i][numPoints+1] = P[i][1];
			A[i][numPoints+2] = P[i][2];
		}

		for (int i = numPoints; i < numPoints+3; i++)
		{
			for (int j = 0; j < numPoints; j++)
			{
				A[i][j] = P[j][i-numPoints];
			}
			A[i][numPoints] = 0;
			A[i][numPoints+1] = 0;
			A[i][numPoints+2] = 0;
		}

		// Setting the right hands side of the system
		double [] b = new double[numPoints + 3];
		for ( int i = 0; i < numPoints; i++)
		{
			b[i] = controlSet2Z[i];
		}
		
		for ( int i = numPoints; i < numPoints + 3; i++)
		{
			b[i] = 0;
		}
		
		double [][] btemp = new double[numPoints + 3][2];
		for (int i = 0; i < numPoints + 3; i ++)
		{
			btemp[i][0] = b[i];
			btemp[i][1] = 1;
		}
		
		Matrix b1 = new Matrix(btemp);
		
		Matrix A1 = new Matrix(A);
		Matrix x1 = A1.solve(b1);
		
		double [][] xtemp = x1.getArray();

		double [] x = new double[numPoints + 3];
		for (int i = 0; i < numPoints + 3; i ++)
		{
			x[i] = xtemp[i][0];
		}

		double [] w = new double[numPoints];

		for ( int i = 0; i < numPoints; i++)
		{
			w[i] = x[i];
		}

		findBE(C, w);
	}
	
	// basis function for planar data X^2*log(X)
	private double basisFunct( double x )
	{
		double bf;	

		if ( x == 0 )
		{
			bf = 0;
		}
		else
		{
			bf = x*x*Math.log(x);
		}

		return bf;
	}
	
	// calculate the bending energy as W^t * C * W
	private void findBE(double [][] C, double [] x)
	{	
		double [] y = new double[numPoints];
		double sum;	
		
		for ( int i = 0; i < numPoints; i++ )
		{
			sum = 0;
			for ( int j = 0; j < numPoints; j++ )
			{
				sum += C[i][j] * x[j]; 
			}	
			y[i] = sum;
		}

		sum = 0;
		for ( int i = 0; i < numPoints; i++ )
		{
			sum += x[i] * y[i];
		}

		BE = sum;
	}
	
	public double getBE()
	{
		return BE;
	}
}
